from maggy.optimizer.abstractoptimizer import AbstractOptimizer
from maggy.optimizer.randomsearch import RandomSearch
from maggy.optimizer.asha import Asha
from maggy.optimizer.singlerun import SingleRun